#ifndef MAKEDEPEND
 #include <cstdio>
 #include <cctype>
 #include <cstdlib>
 #include <string>
 #include <cstdarg>
#endif

#include "disasm.h"

using namespace std;

static const char *regs[] = 
{
	/* 1             4              7     8       9   10   11   12   13   */
	"DH","DL","CH","CL","BH","BL","[BX]","AL",   "DX","CX","BX","SP","AL&FLAGS"
};
static const char opgroup[][5] =
{
	/* 1   2      3     4      5    6     7       8     9      10  */
	"ADD","ADC","SUB","SBB", "AND","XOR","OR",  "CMP", "MOV", "RST",
	"ROL","ROR","RCL","RCR", "SHL","SAR","SWAP","SHR", "STOP","HALT",
	"NOP","INC","DEC","PUSH","POP","STI","CLI", "CALL","RET", "IRET",
	"??", "JMP","DAA","NOT", "STC","CMC","JNZ", "JZ",  "JNC", "JC",
	"RET" /* conditional */
};

#define HALT       20
#define JMP        32
#define STOP       19
#define CALL       28
#define RET        29
#define IRET       30
#define JNZ        37
#define JZ         38
#define JNC        39
#define JC         40

static const unsigned COMMENT_POS = 22;

static const char *ops[256] =
{
/*00*/
	"c21",        "c9\tr9,2",    "c9\t[r9],r8",    "c22\tr9",  "c22\tr1", "c23\tr1", "c9\tr1,1", "c11\tr8,/1",
	"c9\t[2],r12","c1\tr11,r9",  "c9\tr8,[r9]",    "c23\tr9",  "c22\tr2", "c23\tr2", "c9\tr2,1", "c12\tr8,/1",
	"c19 1",      "c9\tr10,2",   "c9\t[r10],r8",   "c22\tr10", "c22\tr3", "c23\tr3", "c9\tr3,1", "c13\tr8,/1",
	"c32\td",     "c1\tr11,r10", "c9\tr8,[r10]",   "c23\tr10", "c22\tr4", "c23\tr4", "c9\tr4,1", "c14\tr8,/1",
	"c37\td",     "c9\tr11,2",   "c9\t[r11++],r8", "c22\tr11", "c22\tr5", "c23\tr5", "c9\tr5,1", "c33",
	"c38\td",     "c1\tr11,r11", "c9\tr8,[r11++]", "c23\tr11", "c22\tr6", "c23\tr6", "c9\tr6,1", "c34\tr8",
	"c39\td",     "c9\tr12,2",   "c9\t[r11--],r8", "c22\tr12", "c22\tr7", "c23\tr7", "c9\tr7,1", "c35",
	"c40\td",     "c1\tr11,r12", "c9\tr8,[r11--]", "c23\tr12", "c22\tr8", "c23\tr8", "c9\tr8,1", "c36",
	
/*40*/
	"c9\tr1,r1", "c9\tr1,r2",   "c9\tr1,r3",   "c9\tr1,r4", "c9\tr1,r5", "c9\tr1,r6", "c9\tr1,r7", "c9\tr1,r8",
	"c9\tr2,r1", "c9\tr2,r2",   "c9\tr2,r3",   "c9\tr2,r4", "c9\tr2,r5", "c9\tr2,r6", "c9\tr2,r7", "c9\tr2,r8",
	"c9\tr3,r1", "c9\tr3,r2",   "c9\tr3,r3",   "c9\tr3,r4", "c9\tr3,r5", "c9\tr3,r6", "c9\tr3,r7", "c9\tr3,r8",
	"c9\tr4,r1", "c9\tr4,r2",   "c9\tr4,r3",   "c9\tr4,r4", "c9\tr4,r5", "c9\tr4,r6", "c9\tr4,r7", "c9\tr4,r8",
	"c9\tr5,r1", "c9\tr5,r2",   "c9\tr5,r3",   "c9\tr5,r4", "c9\tr5,r5", "c9\tr5,r6", "c9\tr5,r7", "c9\tr5,r8",
	"c9\tr6,r1", "c9\tr6,r2",   "c9\tr6,r3",   "c9\tr6,r4", "c9\tr6,r5", "c9\tr6,r6", "c9\tr6,r7", "c9\tr6,r8",
	"c9\tr7,r1", "c9\tr7,r2",   "c9\tr7,r3",   "c9\tr7,r4", "c9\tr7,r5", "c9\tr7,r6", "c20",       "c9\tr7,r8",
	"c9\tr8,r1", "c9\tr8,r2",   "c9\tr8,r3",   "c9\tr8,r4", "c9\tr8,r5", "c9\tr8,r6", "c9\tr8,r7", "c9\tr8,r8",
	
/*80*/
	"c1\tr8,r1", "c1\tr8,r2",   "c1\tr8,r3",   "c1\tr8,r4", "c1\tr8,r5", "c1\tr8,r6", "c1\tr8,r7", "c1\tr8,r8",
	"c2\tr8,r1", "c2\tr8,r2",   "c2\tr8,r3",   "c2\tr8,r4", "c2\tr8,r5", "c2\tr8,r6", "c2\tr8,r7", "c2\tr8,r8",
	"c3\tr8,r1", "c3\tr8,r2",   "c3\tr8,r3",   "c3\tr8,r4", "c3\tr8,r5", "c3\tr8,r6", "c3\tr8,r7", "c3\tr8,r8",
	"c4\tr8,r1", "c4\tr8,r2",   "c4\tr8,r3",   "c4\tr8,r4", "c4\tr8,r5", "c4\tr8,r6", "c4\tr8,r7", "c4\tr8,r8",
	"c5\tr8,r1", "c5\tr8,r2",   "c5\tr8,r3",   "c5\tr8,r4", "c5\tr8,r5", "c5\tr8,r6", "c5\tr8,r7", "c5\tr8,r8",
	"c6\tr8,r1", "c6\tr8,r2",   "c6\tr8,r3",   "c6\tr8,r4", "c6\tr8,r5", "c6\tr8,r6", "c6\tr8,r7", "c6\tr8,r8",
	"c7\tr8,r1", "c7\tr8,r2",   "c7\tr8,r3",   "c7\tr8,r4", "c7\tr8,r5", "c7\tr8,r6", "c7\tr8,r7", "c7\tr8,r8",
	"c8\tr8,r1", "c8\tr8,r2",   "c8\tr8,r3",   "c8\tr8,r4", "c8\tr8,r5", "c8\tr8,r6", "c8\tr8,r7", "c8\tr8,r8",
	
/*C0*/
	"NZ?c41",   "c25\tr9",   "c37\t2",  "c32\t2",    "NZ?c28\t2", "c24\tr9",   "c1\tr8,1",  "c10\t/00",
	"Z?c41",    "c29",       "c38\t2",  "g",         "Z?c28\t2",  "c28\t2",    "c2\tr8,1",  "c10\t/08",
/*D0*/
	"NC?c41",   "c25\tr10",  "c39\t2",  "c31\t!",    "NC?c28\t2", "c24\tr10",  "c3\tr8,1",  "c10\t/10",
	"C?c41",    "c30",       "c40\t2",  "c31\t!",    "C?c28\t2",  "c31\t!",    "c4\tr8,1",  "c10\t/18",
/*E0*/
	"c9\t[$FF\\1],r8", "c25\tr11",   "c9\t[$FF\\/00+r2],r8", "c31\t!", "c31\t!", "c24\tr11", "c5\tr8,1",  "c10\t/20",
	"c1\tr12,+",       "c32\tr7",    "c9\t[2],r8",           "c31\t!", "c31\t!", "c31\t!",   "c6\tr8,1",  "c10\t/28",
	"c9\tr8,[$FF\\1]", "c25\tr13",   "c9\tr8,[r2]",          "c27",    "c31\t!", "c24\tr13", "c7\tr8,1",  "c10\t/30",
	"c9\tr11,r12+",    "c9\tr12,r11","c9\tr8,[2]",           "c26",    "c31\t!", "c31\t!",   "c8\tr8,1",  "c10\t/38",
};
static const char *cbt[256] =
{
/*00 */
	"c11\tr1", "c11\tr2", "c11\tr3", "c11\tr4", "c11\tr5", "c11\tr6", "c11\tr7", "c11\tr8",
	"c12\tr1", "c12\tr2", "c12\tr3", "c12\tr4", "c12\tr5", "c12\tr6", "c12\tr7", "c12\tr8",
	"c13\tr1", "c13\tr2", "c13\tr3", "c13\tr4", "c13\tr5", "c13\tr6", "c13\tr7", "c13\tr8",
	"c14\tr1", "c14\tr2", "c14\tr3", "c14\tr4", "c14\tr5", "c14\tr6", "c14\tr7", "c14\tr8",
	"c15\tr1", "c15\tr2", "c15\tr3", "c15\tr4", "c15\tr5", "c15\tr6", "c15\tr7", "c15\tr8",
	"c16\tr1", "c16\tr2", "c16\tr3", "c16\tr4", "c16\tr5", "c16\tr6", "c16\tr7", "c16\tr8",
	"c17\tr1", "c17\tr2", "c17\tr3", "c17\tr4", "c17\tr5", "c17\tr6", "c17\tr7", "c17\tr8",
	"c18\tr1", "c18\tr2", "c18\tr3", "c18\tr4", "c18\tr5", "c18\tr6", "c18\tr7", "c18\tr8",
/*40*/
	"^0r1%", "^0r2%", "^0r3%", "^0r4%", "^0r5%", "^0r6%", "^0r7%", "^0r8%",
	"^7r1%", "^7r2%", "^7r3%", "^7r4%", "^7r5%", "^7r6%", "^7r7%", "^7r8%",
	"&6r1%", "&6r2%", "&6r3%", "&6r4%", "&6r5%", "&6r6%", "&6r7%", "&6r8%",
	"|5r1%", "|5r2%", "|5r3%", "|5r4%", "|5r5%", "|5r6%", "|5r7%", "|5r8%",
	"^4r1%", "^4r2%", "^4r3%", "^4r4%", "^4r5%", "^4r6%", "^4r7%", "^4r8%",
	"&3r1%", "&3r2%", "&3r3%", "&3r4%", "&3r5%", "&3r6%", "&3r7%", "&3r8%",
	"|2r1%", "|2r2%", "|2r3%", "|2r4%", "|2r5%", "|2r6%", "|2r7%", "|2r8%",
	"^1r1%", "^1r2%", "^1r3%", "^1r4%", "^1r5%", "^1r6%", "^1r7%", "^1r8%",
/*80*/
	"&0r1%", "&0r2%", "&0r3%", "&0r4%", "&0r5%", "&0r6%", "&0r7%", "&0r8%",
	"&7r1%", "&7r2%", "&7r3%", "&7r4%", "&7r5%", "&7r6%", "&7r7%", "&7r8%",
	"|6r1%", "|6r2%", "|6r3%", "|6r4%", "|6r5%", "|6r6%", "|6r7%", "|6r8%",
	"^5r1%", "^5r2%", "^5r3%", "^5r4%", "^5r5%", "^5r6%", "^5r7%", "^5r8%",
	"&4r1%", "&4r2%", "&4r3%", "&4r4%", "&4r5%", "&4r6%", "&4r7%", "&4r8%",
	"|3r1%", "|3r2%", "|3r3%", "|3r4%", "|3r5%", "|3r6%", "|3r7%", "|3r8%",
	"^2r1%", "^2r2%", "^2r3%", "^2r4%", "^2r5%", "^2r6%", "^2r7%", "^2r8%",
	"&1r1%", "&1r2%", "&1r3%", "&1r4%", "&1r5%", "&1r6%", "&1r7%", "&1r8%",
/*C0*/
	"|0r1%", "|0r2%", "|0r3%", "|0r4%", "|0r5%", "|0r6%", "|0r7%", "|0r8%",
	"|7r1%", "|7r2%", "|7r3%", "|7r4%", "|7r5%", "|7r6%", "|7r7%", "|7r8%",
	"^6r1%", "^6r2%", "^6r3%", "^6r4%", "^6r5%", "^6r6%", "^6r7%", "^6r8%",
	"&5r1%", "&5r2%", "&5r3%", "&5r4%", "&5r5%", "&5r6%", "&5r7%", "&5r8%",
	"|4r1%", "|4r2%", "|4r3%", "|4r4%", "|4r5%", "|4r6%", "|4r7%", "|4r8%",
	"^3r1%", "^3r2%", "^3r3%", "^3r4%", "^3r5%", "^3r6%", "^3r7%", "^3r8%",
	"&2r1%", "&2r2%", "&2r3%", "&2r4%", "&2r5%", "&2r6%", "&2r7%", "&2r8%",
	"|1r1%", "|1r2%", "|1r3%", "|1r4%", "|1r5%", "|1r6%", "|1r7%", "|1r8%",
};

#ifdef __GNUC__
static void sprintcat(string &dest, const char *fmt, ...)
     __attribute__ ((__format__ (__printf__, 2, 3)));
#endif
static void sprintcat(string &dest, const char *fmt, ...)
{
	char Buf[1024];
	va_list ap;
	va_start(ap, fmt);
	vsprintf(Buf, fmt, ap);
	va_end(ap);
	dest += Buf;
}

unsigned DisAsm(const byte *data, unsigned pos, bool print,
                bool *jumpflag, bool *callflag,
                bool *endflag, unsigned *jumptarg)
{
	byte b1;
	byte pending = 0;
	bool hexflag = true;
	unsigned pp=pos;
	
	const char *m;
	
	string result;
   		
   	b1 = data[pos++];
   	m = ops[b1];
   	
   	while(*m)
   	{
   		if(*m=='c')
   		{
   			int n=0;
   			while(isdigit(*++m))n=n*10+*m-'0';
   			
   			if(print)result += opgroup[n-1];
   			
   			if(print && n==HALT)
   			{
	    		while(result.size() < COMMENT_POS)result += ' ';
	    		result += "; waits irq, saves power";
   			}
   			
   			*jumpflag = (n==JMP || n==JNZ || n==JZ || n==JNC || n==JC);
   			*callflag = (n==CALL);
   			*endflag  = (n==RET || n==IRET || n==JMP);
   		}
   		else if(*m=='!')
   		{
   			if(print)
   			{
    			if(hexflag)result += '$';
    			sprintcat(result, "%02X", b1);
   			}
   			m++;
   		}
   		else if(*m=='&')
   		{
   			pending = *++m;
   			if(print)result += "BTC      ";
   			m++;
   		}
   		else if(*m=='|')
   		{
   			pending = *++m;
   			if(print)result += "BTS      ";
   			m++;
   		}
   		else if(*m=='^')
   		{
   			pending = *++m;
   			if(print)result += "TEST     ";
   			m++;
   		}
   		else if(*m=='%')
   		{
   			if(print)
   			{
   				result += ',';
   				result += (char)pending;
	    	//	while(result.size() < COMMENT_POS)result += ' ';
	    	//	result += "; bit op";
   			}
   			m++;
   		}
   		else if(*m=='r')
   		{
   			int n=0;
   			while(isdigit(*++m))n=n*10+*m-'0';
   			if(print)result += regs[n-1];
   		}
   		else if(*m=='+' && m[1]!='+' && m[1]!=']')
   		{
   			byte b2 = (signed char)data[pos++];
   			if(print)
   			{
   				result += b2&0x80 ? '-' : '+';
   				if(hexflag)result += '$';
   				if(b2&0x80)b2 = -(signed char)b2;
   				sprintcat(result, "%02X", b2);
   			}
   			m++;
   		}
   		else if(*m=='1')
   		{
   			byte b2 = data[pos++];
   			if(print)
   			{
    			if(hexflag)result += '$';
    			sprintcat(result, "%02X", b2);
   			}
   			m++;
   		}
   		else if(*m=='d')
   		{
   			char c = data[pos];
   			unsigned p = pos + c + 1;
   			pos++;
   			
   			/* displacement */
   			
   			if(print)
   			{
    			if(hexflag)result += '$';
    			sprintcat(result, "%04X", p);
   			}
   			m++;
   			
   			*jumptarg = p;
   		}
   		else if(*m=='2')
   		{
   			byte b2 = data[pos++];
   			byte b3 = data[pos++];
   			unsigned d = b2 | ((unsigned)b3<<8);
   			
   			*jumptarg = d;
   			if(print)
   			{
    			if(hexflag)result += '$';
    			sprintcat(result, "%04X", d);
   			}
   			m++;
   		}
   		else if(*m=='/')
   		{	
   			char Buf[16], *s=Buf;
   			int l;
   			for(l=0; isxdigit(*++m) && l<15; l++)
   				*s++=*m;
   			*s=0;
   			if(print)
   			{
    			if(hexflag && strtol(Buf,NULL,16) >= 10)result += '$';
    			result += Buf;
			}
   		}
   		else if(*m=='g')
   		{
   			m = cbt[data[pos++]];
   		}
   		else if(*m=='\t')
   		{
   			if(print)while(result.size()%9)result += ' ';
   			m++;
   		}
   		else if(*m!='\\')
   		{
   			if(print)result += *m;
   			m++;
   		}   			
   		if(*m=='\\')
   		{
   			m++;
   			hexflag=false;
   		}
   		else
   			hexflag=true;
   	}
   	if(print)
   	{
   		unsigned a, b = pos-pp;
   		printf("\33[36m%04X ", pp);
   		for(a=0; a<b; a++)printf("%02X ", data[pp+a]);
   		for(; a<6; a++)printf("   ");
   		printf("%s\33[37m", result.c_str());
   	}
	return pos-pp;
}
