#ifndef __DISASM_H_
#define __DISASM_H_

// Disassembles one instruction from data at position pos.
// The position shall not be >= 0x8000, because it is intended
// to map to REAL Game Boy memory.
// Set print to true if the disassembled instruction wants to be printed.
//
// Parameters:
//       data    -- pointer to Game Boy memory space (0x10000 bytes)
//       pos     -- position to disassemble (should be <0x8000)
//       print   -- if false, this does not output any text (printf).
//
//       jumpflag -- pointer to a bool that shall be set 'true'
//                   if the instruction was a jump command.
//       callflag -- the same for calls.
//       endflag  -- pointer to a bool that shall be set 'true'
//                   if the code flow could not continue to the
//                   next instruction (was ret or reti or uncond jump).
//       jumptarg -- if jumpflag was set to true, then jumptarg
//                   will be set to contain the jump target.
//
// Return value:
//       The size of instruction was (number of bytes).

typedef unsigned char byte;
typedef unsigned int  word;

extern unsigned DisAsm(const byte *data, unsigned pos, bool print,
                       bool *jumpflag, bool *callflag,
                       bool *endflag, unsigned *jumptarg);

extern void compile(const byte *data, unsigned pos, unsigned length, bool enterblock, unsigned endjump=0);
extern bool haslabel(const byte *data, unsigned pos);
extern unsigned getjumpcount(const byte *data, unsigned pos);
extern unsigned getjump(const byte *data, unsigned pos, unsigned ind);
extern void MarkVisited(const byte *data, unsigned pos);
extern bool alreadycompiled(const byte *data, unsigned pos);
extern void RequestCompile(const byte *data, unsigned pos, bool enterblock, bool doendjump=true);

extern unsigned emulate();
extern byte PEEK(const word w);
extern void POKE(const word w, const byte b);
extern void DISABLE();
extern void ENABLE();
extern void OPCODE(byte);   // for unknown opcodes
extern void STOP(byte);     // for very low power sleep mode.
extern void HALT();         // wait for an irq
extern void RETI();         // only reallow irqs after IRET.
extern word PC;

extern void IRQ(byte);      // 1=vblank, 2=lcd, 4=timer, 8=serial, 16=joy

#endif
