#ifndef MAKEDEPEND
 #include <cstdio>
 #include <unistd.h>
#endif

#include "disasm.h"
#include "emuvars.h"

static void remap(void)
{
	ROMs[0] = ROM;
	ROMs[1] = ROM + rombank*0x4000;
	RAMs[0] = RAM + rambank*0x2000;
}

#include "joypad.h"
#include "divreg.h"
#include "timerreg.h"
#include "sound.h"

void DISABLE()
{
}
void ENABLE()
{
}
void OPCODE(byte)
{
	// for unknown opcodes
}
void STOP(byte)
{
    // for very low power sleep mode.
	usleep(1000);
}
void HALT()
{
	usleep(10);
    // wait for an irq
}
void RETI()
{
    // only reallow irqs after IRET
}
void IRQ(byte)
{

	// 1=vblank, 2=lcd, 4=timer, 8=serial, 16=joy
}

byte PEEK(const word w)
{
	if(w < 0x8000) return ROMs[w/0x4000][w&0x3FFF];
	if(w < 0xA000) return VideoRAM[w&0x1FFF];
	if(w < 0xC000) return RAMs[0][w&0x1FFF];
	if(w == 0xFFFF) return InterruptEnable;
	if(w < 0xFE00 || (w>=0xFF80)) return InternalRAM[w&0x1FFF];
	if(w < 0xFEA0) return OAM[w&0xFF];
	switch(w)
	{
		case 0xFF00: return P1.read();
		case 0xFF04: return DIV.read(); 
		case 0xFF05: return TIMA.read();
		case 0xFF06: return TMA.read();
	}
	return 0;
}
void POKE(const word w, const byte b)
{
	if(w < 0x2000)      enableram = b;
	else if(w < 0x4000) rombank = (b?b:1) % numrombanks;
	else if(w < 0x6000) rambank = b % numrambanks;
	else if(w < 0x8000) return;
	else if(w < 0xA000) VideoRAM[w&0x1FFF] = b; // NOTE
	else if(w < 0xC000) { if(enableram)RAMs[0][w&0x1FFF]=b; }
	else if(w == 0xFFFF) InterruptEnable = b;
	else if(w < 0xFE00 || (w >= 0xFF80)) InternalRAM[w&0x1FFF] = b;
	else if(w < 0xFEA0) OAM[w&0xFF] = b;
	else
		switch(w)
		{
			case 0xFF00: P1.write(b); break;
			case 0xFF04: DIV.write(b); break;
			case 0xFF05: TIMA.write(b); break;
			case 0xFF06: TMA.write(b); break;
			case 0xFF07: TAC.write(b); break;
		}
}

word PC;

void emuinit(const byte *d)
{
	ROM = d;
	switch(ROM[0x148])
	{
		case 0: numrombanks = 2; break;
		case 1: numrombanks = 4; break;
		case 2: numrombanks = 8; break;
		case 3: numrombanks = 16; break;
		case 4: numrombanks = 32; break;
		case 5: numrombanks = 64; break;
		case 6: numrombanks = 128; break;
		case 0x52: numrombanks = 72; break;
		case 0x53: numrombanks = 80; break;
		case 0x54: numrombanks = 96; break;
		default: numrombanks = 2; break;
	}
	switch(ROM[0x149])
	{
		case 0: numrambanks = 0; break;
		case 1:
		case 2: numrambanks = 1; break;
		case 3: numrambanks = 4; break;
		default:
			numrambanks = 0;
	}
	switch(ROM[0x147])
	{
		case 3:
		case 6:
		case 9:
		case 0xFF: battery = true;
		default: battery = false;
	}
	
	if(RAM)delete[] RAM;
	RAM = new byte[0x2000 * numrambanks];
	
	SuperGameBoy = ROM[0x14B] == 0x33;
	enableram = false;
	rombank=1;
	rambank=0;
	remap();
}
