#ifndef gbemu_TIMER_H
#define gbemu_TIMER_H

#include "threads.h"

#ifndef MAKEDEPEND
#include <vector>
#include <unistd.h>
#endif

#define SMALLEST_USLEEP 10000

class timer
{
	static const unsigned disabled=0;
	friend class timers;
	mutex mut;
	unsigned rate;
	double pos;
	void (*fun)();
public:
	timer(unsigned r, void (*f)());
	unsigned getrate() const
	{
		return rate;
	}
	void disable()
	{
		mut.lock();
		rate = disabled;
		mut.unlock();
	}
	void changerate(unsigned r)
	{
		mut.lock();
		rate = r;
		mut.unlock();
	}	
};
static void *TimerUpdater(void *);
class timers
{
	friend class timer;
	unsigned resolution;
	mutex mut;
	vector <timer *> tim;
	bool active;
	void activate()
	{
		active=true;
		ThreadManager.add(TimerUpdater, NULL);
	}
public:
	timers(unsigned res) :resolution(res), active(false) { }
	~timers()
	{
	}
	void add(timer *t)
	{
		if(!active)activate();
		mut.lock();
		tim.push_back(t);
		mut.unlock();
	}
	void run()
	{
		for(;;)
		{
			usleep(resolution);
			mut.lock();
			unsigned a, b=tim.size();
			for(a=0; a<b; a++)
			{
				timer *c = tim[a];
				if(c->rate)
				{
					c->pos += c->rate * 1E-6 * resolution;
					while(c->pos >= 1.0)
					{
						c->fun();
						--c->pos;
					}
				}
			}
			mut.unlock();
		}
	}
} TimerManager(SMALLEST_USLEEP);
static void *TimerUpdater(void *)
{
	TimerManager.run();
	return NULL;
}

timer::timer(unsigned r, void (*f)()) : rate(r), pos(0.0), fun(f)
{
	TimerManager.add(this);
}

#endif

